/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel.data;

import java.util.Arrays;
import java.util.List;

import org.eobjects.metamodel.query.SelectItem;
import org.eobjects.metamodel.schema.Column;
import org.eobjects.metamodel.util.BaseObject;

/**
 * Default Row implementation. Holds values in memory.
 * 
 * @author Kasper Sørensen
 */
public final class DefaultRow extends BaseObject implements Cloneable, Row {

    private static final long serialVersionUID = 1L;

    private final List<SelectItem> _items;
    private final Object[] _values;
    private final Style[] _styles;

    /**
     * Constructs a row from an array of SelectItems and an array of
     * corresponding values
     * 
     * @param items
     *            the array of SelectItems
     * @param values
     *            the array of values
     */
    public DefaultRow(SelectItem[] items, Object[] values) {
        this(Arrays.asList(items), values, null);
    }

    /**
     * Constructs a row from an array of SelectItems and an array of
     * corresponding values
     * 
     * @param items
     *            the array of SelectItems
     * @param values
     *            the array of values
     * @param styles
     *            an optional array of styles
     */
    public DefaultRow(SelectItem[] items, Object[] values, Style[] styles) {
        this(Arrays.asList(items), values, styles);
    }

    /**
     * Constructs a row from a list of SelectItems and an array of corresponding
     * values
     * 
     * @param items
     *            the list of SelectItems
     * @param values
     *            the array of values
     */
    public DefaultRow(List<SelectItem> items, Object[] values) {
        this(items, values, null);
    }

    /**
     * Constructs a row from a list of SelectItems and an array of corresponding
     * values
     * 
     * @param items
     *            the list of SelectItems
     * @param values
     *            the array of values
     * @param styles
     *            an optional array of styles
     */
    public DefaultRow(List<SelectItem> items, Object[] values, Style[] styles) {
        if (items == null) {
            throw new IllegalArgumentException("Select items cannot be null");
        }
        if (values == null) {
            throw new IllegalArgumentException("Select items cannot be null");
        }
        if (items.size() != values.length) {
            throw new IllegalArgumentException("SelectItems length and values length must be equal. " + items.size()
                    + " select items present and encountered these values: " + Arrays.toString(values));
        }
        if (styles != null) {
            if (items.size() != styles.length) {
                throw new IllegalArgumentException("SelectItems length and styles length must be equal. "
                        + items.size() + " select items present and encountered these values: "
                        + Arrays.toString(styles));
            }
            boolean entirelyNoStyle = true;
            for (int i = 0; i < styles.length; i++) {
                if (styles[i] == null) {
                    throw new IllegalArgumentException("Elements in the style array cannot be null");
                }
                if (entirelyNoStyle && !Style.NO_STYLE.equals(styles[i])) {
                    entirelyNoStyle = false;
                }
            }

            if (entirelyNoStyle) {
                // no need to reference any styles
                styles = null;
            }
        }
        _items = items;
        _values = values;
        _styles = styles;
    }

    @Override
    public Object getValue(SelectItem item) {
        if (item == null) {
            return null;
        }

        int i = 0;
        for (SelectItem selectItem : _items) {
            if (item == selectItem) {
                return _values[i];
            }
            i++;
        }

        i = 0;
        for (SelectItem selectItem : _items) {
            if (item.equalsIgnoreAlias(selectItem, true)) {
                return _values[i];
            }
            i++;
        }

        i = 0;
        for (SelectItem selectItem : _items) {
            if (item.equalsIgnoreAlias(selectItem)) {
                return _values[i];
            }
            i++;
        }

        return null;
    }

    @Override
    public Object getValue(Column column) {
        return getValue(indexOf(column));
    }

    @Override
    public int indexOf(SelectItem item) {
        if (item == null) {
            return -1;
        }
        int i = 0;
        for (SelectItem selectItem : _items) {
            if (item.equalsIgnoreAlias(selectItem)) {
                return i;
            }
            i++;
        }
        return -1;
    }

    @Override
    public int indexOf(Column column) {
        if (column == null) {
            return -1;
        }
        SelectItem selectItem = new SelectItem(column);
        return indexOf(selectItem);
    }

    @Override
    public Object getValue(int index) throws ArrayIndexOutOfBoundsException {
        return _values[index];
    }

    @Override
    public SelectItem[] getSelectItems() {
        return _items.toArray(new SelectItem[_items.size()]);
    }

    @Override
    public Object[] getValues() {
        return _values;
    }

    @Override
    public String toString() {
        return "Row[values=" + Arrays.toString(_values) + "]";
    }

    @Override
    protected void decorateIdentity(List<Object> identifiers) {
        identifiers.add(_items);
        identifiers.add(_values);
        identifiers.add(_styles);
    }

    @Override
    protected DefaultRow clone() {
        return new DefaultRow(_items, _values, _styles);
    }

    @Override
    public Row getSubSelection(final SelectItem[] selectItems) {
        final Object[] values = new Object[selectItems.length];
        final Style[] styles = new Style[selectItems.length];
        for (int i = 0; i < selectItems.length; i++) {
            SelectItem selectItem = selectItems[i];

            if (selectItem.getSubQuerySelectItem() != null) {
                values[i] = getValue(selectItem.getSubQuerySelectItem());
                styles[i] = getStyle(selectItem.getSubQuerySelectItem());
                if (values[i] == null) {
                    values[i] = getValue(selectItem);
                    styles[i] = getStyle(selectItem);
                }
            } else {
                values[i] = getValue(selectItem);
                styles[i] = getStyle(selectItem);
            }
        }
        return new DefaultRow(selectItems, values, styles);
    }

    @Override
    public Style getStyle(SelectItem item) {
        if (_styles == null) {
            return Style.NO_STYLE;
        }
        return _styles[indexOf(item)];
    }

    @Override
    public Style getStyle(Column column) {
        if (_styles == null) {
            return Style.NO_STYLE;
        }
        return _styles[indexOf(column)];
    }

    @Override
    public Style getStyle(int index) throws IndexOutOfBoundsException {
        if (_styles == null) {
            return Style.NO_STYLE;
        }
        return _styles[index];
    }

    @Override
    public int size() {
        return _items.size();
    }
}