/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel.delete;

import org.eobjects.metamodel.DataContext;
import org.eobjects.metamodel.MetaModelException;
import org.eobjects.metamodel.query.FilterItem;
import org.eobjects.metamodel.query.builder.FilterBuilder;
import org.eobjects.metamodel.schema.Column;
import org.eobjects.metamodel.schema.Table;
import org.eobjects.metamodel.update.RowUpdationBuilder;

/**
 * Builder object for row deletions in a {@link Table}.
 * 
 * @author Kasper Sørensen
 */
public interface RowDeletionBuilder {

    /**
     * Defines a where item to set as a criteria for this delete
     * 
     * @param column
     *            a column to apply a criteria for
     * @return a {@link FilterBuilder} for further building the where item
     */
    public FilterBuilder<RowDeletionBuilder> where(Column column);

    /**
     * Defines a where item to set as a criteria for this delete
     * 
     * @param columnName
     *            the name of the colum to which the criteria will be applied
     * @return a {@link FilterBuilder} for further building the where item
     */
    public FilterBuilder<RowDeletionBuilder> where(String columnName);

    /**
     * Applies where items to set criteria for this delete
     * 
     * @param filterItems
     *            the where items to set
     * @return the {@link RowUpdationBuilder} itself, for further building of
     *         the update
     */
    public RowDeletionBuilder where(FilterItem... filterItems);

    /**
     * Applies where items to set criteria for this delete
     * 
     * @param filterItems
     *            the where items to set
     * @return the {@link RowUpdationBuilder} itself, for further building of
     *         the update
     */
    public RowDeletionBuilder where(Iterable<FilterItem> filterItems);

    /**
     * Gets the table that this delete statement pertains to.
     * 
     * @return the table that this delete statement pertains to.
     */
    public Table getTable();

    /**
     * Gets a SQL representation of this delete operation. Note that the
     * generated SQL is dialect agnostic, so it is not accurately the same as
     * what will be passed to a potential backing database.
     * 
     * @return a SQL representation of this delete operation.
     */
    public String toSql();

    /**
     * Commits the row deletion operation. This operation will delete rows in
     * the {@link DataContext}.
     * 
     * @throws MetaModelException
     *             if the operation was rejected
     */
    public void execute() throws MetaModelException;
}
