/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel.query.builder;

import java.util.List;

import org.eobjects.metamodel.DataContext;
import org.eobjects.metamodel.data.DataSet;
import org.eobjects.metamodel.query.FilterItem;
import org.eobjects.metamodel.query.FromItem;
import org.eobjects.metamodel.query.FunctionType;
import org.eobjects.metamodel.query.Query;
import org.eobjects.metamodel.query.SelectItem;
import org.eobjects.metamodel.schema.Column;
import org.eobjects.metamodel.schema.Table;
import org.eobjects.metamodel.util.BaseObject;

final class GroupedQueryBuilderImpl extends BaseObject implements GroupedQueryBuilder {

    private Query query;
    private DataContext dataContext;

    public GroupedQueryBuilderImpl(DataContext dataContext, Query query) {
        if (query == null) {
            throw new IllegalArgumentException("query cannot be null");
        }
        this.dataContext = dataContext;
        this.query = query;
    }

    @Override
    public ColumnSelectBuilder<GroupedQueryBuilder> select(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return new ColumnSelectBuilderImpl(column, query, this);
    }

    @Override
    public FunctionSelectBuilder<GroupedQueryBuilder> select(FunctionType function, Column column) {
        if (function == null) {
            throw new IllegalArgumentException("function cannot be null");
        }
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return new FunctionSelectBuilderImpl(function, column, query, this);
    }

    @Override
    public SatisfiedQueryBuilder<GroupedQueryBuilder> where(FilterItem... filters) {
        query.where(filters);
        return this;
    }
    
    @Override
    public SatisfiedQueryBuilder<GroupedQueryBuilder> where(Iterable<FilterItem> filters) {
        query.where(filters);
        return this;
    }

    @Override
    public ColumnSelectBuilder<GroupedQueryBuilder> select(String columnName) {
        Column column = findColumn(columnName);
        return select(column);
    }

    @Override
    public CountSelectBuilder<GroupedQueryBuilder> selectCount() {
        return new CountSelectBuilderImpl(query, this);
    }

    @Override
    public SatisfiedSelectBuilder<GroupedQueryBuilder> select(Column... columns) {
        if (columns == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        query.select(columns);
        return new SatisfiedSelectBuilderImpl(this);
    }

    @Override
    public WhereBuilder<GroupedQueryBuilder> where(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return new WhereBuilderImpl(column, query, this);
    }

    @Override
    public WhereBuilder<GroupedQueryBuilder> where(String columnName) {
        Column column = findColumn(columnName);
        return where(column);
    }

    @Override
    public Column findColumn(String columnName) throws IllegalArgumentException {
        if (columnName == null) {
            throw new IllegalArgumentException("columnName cannot be null");
        }

        List<FromItem> fromItems = query.getFromClause().getItems();
        List<SelectItem> selectItems = query.getSelectClause().getItems();

        int dotIndex = columnName.indexOf('.');
        if (dotIndex != -1) {
            // check aliases of from items
            String aliasPart = columnName.substring(0, dotIndex);
            String columnPart = columnName.substring(dotIndex + 1);

            for (FromItem fromItem : fromItems) {
                Column column = null;
                column = findColumnInAliasedTable(column, fromItem, aliasPart, columnPart);
                if (column != null) {
                    return column;
                }
            }
        }

        // check columns already in select clause
        for (SelectItem item : selectItems) {
            Column column = item.getColumn();
            if (column != null) {
                if (columnName.equals(column.getName())) {
                    return column;
                }
            }
        }

        for (FromItem fromItem : fromItems) {
            Table table = fromItem.getTable();
            if (table != null) {
                Column column = table.getColumnByName(columnName);
                if (column != null) {
                    return column;
                }
            }
        }

        Column column = dataContext.getColumnByQualifiedLabel(columnName);
        if (column != null) {
            return column;
        }

        throw new IllegalArgumentException("Could not find column: " + columnName);
    }

    private Column findColumnInAliasedTable(Column column, FromItem fromItem, String aliasPart, String columnPart) {
        if (column != null) {
            // ensure that if the column has already been found, return it
            return column;
        }

        Table table = fromItem.getTable();
        if (table != null) {
            String alias = fromItem.getAlias();
            if (alias != null && alias.equals(aliasPart)) {
                column = table.getColumnByName(columnPart);
            }
        } else {
            FromItem leftSide = fromItem.getLeftSide();
            column = findColumnInAliasedTable(column, leftSide, aliasPart, columnPart);
            FromItem rightSide = fromItem.getRightSide();
            column = findColumnInAliasedTable(column, rightSide, aliasPart, columnPart);
            if (column != null) {
                Query subQuery = fromItem.getSubQuery();
                if (subQuery != null) {
                    List<FromItem> items = subQuery.getFromClause().getItems();
                    for (FromItem subQueryFromItem : items) {
                        column = findColumnInAliasedTable(column, subQueryFromItem, aliasPart, columnPart);
                    }
                }
            }
        }

        return column;
    }

    @Override
    public SatisfiedOrderByBuilder<GroupedQueryBuilder> orderBy(String columnName) {
        return orderBy(findColumn(columnName));
    }

    @Override
    public SatisfiedOrderByBuilder<GroupedQueryBuilder> orderBy(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return new SatisfiedOrderByBuilderImpl(column, query, this);
    }

    @Override
    public SatisfiedOrderByBuilder<GroupedQueryBuilder> orderBy(FunctionType function, Column column) {
        if (function == null) {
            throw new IllegalArgumentException("function cannot be null");
        }
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return new SatisfiedOrderByBuilderImpl(function, column, query, this);
    }

    @Override
    public GroupedQueryBuilder groupBy(Column column) {
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        query.groupBy(column);
        return this;
    }

    @Override
    public GroupedQueryBuilder groupBy(String columnName) {
        Column column = findColumn(columnName);
        return groupBy(column);
    }

    @Override
    public GroupedQueryBuilder groupBy(Column... columns) {
        if (columns == null) {
            throw new IllegalArgumentException("columns cannot be null");
        }
        query.groupBy(columns);
        return this;
    }

    @Override
    public HavingBuilder having(FunctionType function, Column column) {
        if (function == null) {
            throw new IllegalArgumentException("function cannot be null");
        }
        if (column == null) {
            throw new IllegalArgumentException("column cannot be null");
        }
        return new HavingBuilderImpl(function, column, query, this);
    }

    @Override
    public String toString() {
        return query.toSql();
    }

    @Override
    public Query toQuery() {
        return query.clone();
    }

    @Override
    public DataSet execute() {
        return dataContext.executeQuery(query);
    }

    @Override
    protected void decorateIdentity(List<Object> identifiers) {
        identifiers.add(query);
    }
}