/**
 * eobjects.org MetaModel
 * Copyright (C) 2010 eobjects.org
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution; if not, write to:
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301  USA
 */
package org.eobjects.metamodel.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.Flushable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Various helper methods for handling files
 */
public final class FileHelper {

    private final static Logger logger = LoggerFactory
            .getLogger(FileHelper.class);

    public static final String UTF_8_ENCODING = "UTF-8";
    public static final String UTF_16_ENCODING = "UTF-16";
    public static final String US_ASCII_ENCODING = "US-ASCII";
    public static final String ISO_8859_1_ENCODING = "ISO_8859_1";
    public static final String DEFAULT_ENCODING = UTF_8_ENCODING;

    private FileHelper() {
        // prevent instantiation
    }

    public static File getTempDir() {
        File result = null;
        String tmpDirPath = System.getenv("java.io.tmpdir");
        if (tmpDirPath != null && !"".equals(tmpDirPath)) {
            result = new File(tmpDirPath);
        } else {
            logger.debug("Could not determine tmpdir by using environment variable.");
            try {
                File file = File.createTempFile("foo", "bar");
                result = file.getParentFile();
                if (!file.delete()) {
                    logger.warn("Could not delete temp file '{}'",
                            file.getAbsolutePath());
                }
            } catch (IOException e) {
                logger.error(
                        "Could not create tempFile in order to find temporary dir",
                        e);
                result = new File("metamodel.tmp.dir");
                if (!result.mkdir()) {
                    throw new IllegalStateException(
                            "Could not create directory for temporary files: "
                                    + result.getName());
                }
                result.deleteOnExit();
            }
        }
        if (logger.isInfoEnabled()) {
            logger.info("Using '{}' as tmpdir.", result.getAbsolutePath());
        }
        return result;
    }

    public static Writer getWriter(File file, String encoding, boolean append) {
        boolean insertBom = !append;
        return getWriter(file, encoding, append, insertBom);
    }

    public static Writer getWriter(File file, String encoding, boolean append,
            boolean insertBom) {
        if (append && insertBom) {
            throw new IllegalArgumentException(
                    "Can not insert BOM into appending writer");
        }
        try {
            FileOutputStream fileOutputStream = new FileOutputStream(file,
                    append);
            if (insertBom) {
                Writer writer = new UnicodeWriter(fileOutputStream, encoding);
                return writer;
            } else {
                Writer writer = new OutputStreamWriter(fileOutputStream,
                        encoding);
                return writer;
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static Writer getWriter(File file, String encoding) {
        return getWriter(file, encoding, false);
    }

    public static void writeStringAsFile(File file, String string,
            String encoding) {
        try {
            BufferedWriter bw = getBufferedWriter(file, encoding);
            bw.write(string);
            bw.close();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static Reader getReader(File file, String encoding) {
        InputStream inputStream;
        try {
            inputStream = new FileInputStream(file);
            if (encoding == null || encoding.toLowerCase().indexOf("utf") != -1) {
                byte bom[] = new byte[4];
                int unread;

                // auto-detect byte-order-mark
                PushbackInputStream pushbackInputStream = new PushbackInputStream(
                        inputStream, bom.length);
                int n = pushbackInputStream.read(bom, 0, bom.length);

                // Read ahead four bytes and check for BOM marks.
                if ((bom[0] == (byte) 0xEF) && (bom[1] == (byte) 0xBB)
                        && (bom[2] == (byte) 0xBF)) {
                    encoding = "UTF-8";
                    unread = n - 3;
                } else if ((bom[0] == (byte) 0xFE) && (bom[1] == (byte) 0xFF)) {
                    encoding = "UTF-16BE";
                    unread = n - 2;
                } else if ((bom[0] == (byte) 0xFF) && (bom[1] == (byte) 0xFE)) {
                    encoding = "UTF-16LE";
                    unread = n - 2;
                } else if ((bom[0] == (byte) 0x00) && (bom[1] == (byte) 0x00)
                        && (bom[2] == (byte) 0xFE) && (bom[3] == (byte) 0xFF)) {
                    encoding = "UTF-32BE";
                    unread = n - 4;
                } else if ((bom[0] == (byte) 0xFF) && (bom[1] == (byte) 0xFE)
                        && (bom[2] == (byte) 0x00) && (bom[3] == (byte) 0x00)) {
                    encoding = "UTF-32LE";
                    unread = n - 4;
                } else {
                    unread = n;
                }

                if (unread > 0) {
                    pushbackInputStream.unread(bom, (n - unread), unread);
                } else if (unread < -1) {
                    pushbackInputStream.unread(bom, 0, 0);
                }

                inputStream = pushbackInputStream;
            }

            final InputStreamReader inputStreamReader;
            if (encoding == null) {
                inputStreamReader = new InputStreamReader(inputStream);
            } else {
                inputStreamReader = new InputStreamReader(inputStream, encoding);
            }
            return inputStreamReader;
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static String readFileAsString(File file, String encoding) {
        try {
            BufferedReader br = getBufferedReader(file, encoding);
            StringBuilder sb = new StringBuilder();
            boolean firstLine = true;
            for (String line = br.readLine(); line != null; line = br
                    .readLine()) {
                if (firstLine) {
                    firstLine = false;
                } else {
                    sb.append('\n');
                }
                sb.append(line);
            }
            br.close();
            return sb.toString();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static void safeClose(Object... objects) {
        boolean debugEnabled = logger.isDebugEnabled();

        if (objects == null || objects.length == 0) {
            logger.info(
                    "safeClose(...) was invoked with null or empty array: {}",
                    objects);
            return;
        }

        for (Object obj : objects) {
            if (obj != null) {
                if (debugEnabled) {
                    logger.debug("Trying to safely close {}", obj);
                }

                if (obj instanceof Flushable) {
                    try {
                        ((Flushable) obj).flush();
                    } catch (Exception e) {
                        if (debugEnabled) {
                            logger.debug("Flushing Flushable failed", e);
                        }
                    }
                }

                if (obj instanceof Closeable) {
                    try {
                        ((Closeable) obj).close();
                    } catch (IOException e) {
                        if (debugEnabled) {
                            logger.debug("Closing Closeable failed", e);
                        }
                    }
                } else if (obj instanceof Connection) {
                    try {
                        ((Connection) obj).close();
                    } catch (Exception e) {
                        if (debugEnabled) {
                            logger.debug("Closing Connection failed", e);
                        }
                    }
                } else if (obj instanceof Statement) {
                    try {
                        ((Statement) obj).close();
                    } catch (Exception e) {
                        if (debugEnabled) {
                            logger.debug("Closing Statement failed", e);
                        }
                    }
                } else if (obj instanceof ResultSet) {
                    try {
                        ((ResultSet) obj).close();
                    } catch (Exception e) {
                        if (debugEnabled) {
                            logger.debug("Closing ResultSet failed", e);
                        }
                    }
                } else {
                    logger.info("obj was neither Closeable, Connection, Statement or ResultSet.");

                    try {
                        Method method = obj.getClass().getMethod("close",
                                new Class[0]);
                        if (method == null) {
                            logger.info("obj did not have a close() method, ignoring");
                        } else {
                            method.setAccessible(true);
                            method.invoke(obj);
                        }
                    } catch (InvocationTargetException e) {
                        logger.warn(
                                "Invoking close() by reflection threw exception",
                                e);
                    } catch (Exception e) {
                        logger.warn("Could not invoke close() by reflection", e);
                    }
                }
            }

        }
    }

    public static BufferedWriter getBufferedWriter(File file, String encoding) {
        return new BufferedWriter(getWriter(file, encoding));
    }

    public static BufferedReader getBufferedReader(File file, String encoding) {
        return new BufferedReader(getReader(file, encoding));
    }

    public static Reader getReader(File file) {
        return getReader(file, DEFAULT_ENCODING);
    }

    public static String readFileAsString(File file) {
        return readFileAsString(file, DEFAULT_ENCODING);
    }

    public static BufferedWriter getBufferedWriter(File file) {
        return getBufferedWriter(file, DEFAULT_ENCODING);
    }

    public static Writer getWriter(File file) {
        return getWriter(file, DEFAULT_ENCODING);
    }

    public static void writeStringAsFile(File file, String string) {
        writeStringAsFile(file, string, DEFAULT_ENCODING);
    }

    public static BufferedReader getBufferedReader(File file) {
        return getBufferedReader(file, DEFAULT_ENCODING);
    }

    public static void copy(File from, File to) {
        assert from.exists();
        FileInputStream fromStream = null;
        FileOutputStream toStream = null;
        try {
            fromStream = new FileInputStream(from);
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
        try {
            toStream = new FileOutputStream(to);
            copy(fromStream, toStream);
        } catch (IOException e) {
            throw new IllegalStateException(e);
        } finally {
            safeClose(fromStream, toStream);
        }
    }

    public static void copy(InputStream fromStream, OutputStream toStream)
            throws IOException {
        byte[] buffer = new byte[1024 * 32];
        for (int read = fromStream.read(buffer); read != -1; read = fromStream
                .read(buffer)) {
            toStream.write(buffer, 0, read);
        }
    }
}